import code

import matplotlib
import numpy as np
from matplotlib.ticker import FormatStrFormatter

import hex3_chen_old as hex3_chen

from simulation.simulation import main

matplotlib.use('Qt5agg')
import matplotlib.pyplot as plt


def full_run(disturb_table, dist_idx, meas_config):
    runs = []
    import pandas as pd
    hex3_chen.Ti_max = 135
    samples = 500
    if meas_config == 't only': # Run gc only for t only runs, because not needed to repeat em.
        controllers = ['uc', 'uc2', 'gc']
    else:
        controllers = ['uc', 'uc2']

    for mode in controllers:
        for MS in range(1, 5):
            MS = -MS
            for noise in [True, False]:
                try:
                    # GP controller
                    do_plot = True
                    np.random.seed(21)

                    if mode == 'uc':
                        from simulation.predictor_constrained_uopt import InputConstrainedController

                        # train_data = f'optimal_u_constrained\\datasets\\test_u_constrained_prediction{12500}.csv'
                        train_data = f'optimal_u_constrained\\datasets\\train_u_constrained_prediction{samples}.csv'
                        controller = InputConstrainedController(train_data, meas_set=MS, noise=noise, meas_config=meas_config)
                    elif mode == 'uc2':
                        from simulation.predictor_constrained_uopt_v2 import InputConstrainedController

                        # train_data = f'optimal_u_constrained\\datasets\\test_u_constrained_prediction{12500}.csv'
                        train_data = f'optimal_u_constrained\\datasets\\train_u_constrained_prediction{samples}.csv'
                        # train_data = f'optimal_u\\datasets\\train_u_prediction{500}.csv'
                        controller = InputConstrainedController(train_data, meas_set=MS, noise=noise, meas_config=meas_config)
                    elif mode == 'gc':
                        from simulation.predictor_constrained_gradient import GradientConstrainedController
                        train_data = f'gradient\\datasets\\new_train_gradient{samples}.csv'
                        # train_data = f'gradient_constrained\\datasets\\train_constrained_gradient{500}.csv'
                        controller = GradientConstrainedController(train_data, meas_set=MS, noise=noise)
                    else:
                        raise Exception(f'unknown mode {mode}')

                    result = main(MS, do_plot, controller, disturb_table, noise)
                    entry = {
                        'data': result,
                        'total loss': result['total loss'],
                        'negative loss': result['negative loss'],
                        # 'max step loss': result['max step loss'],
                        'mode': mode,
                        'MS': MS,
                        'noise': noise,
                    }
                    print(entry['MS'], entry['mode'], entry['noise'])
                    runs.append(entry)

                    if do_plot:

                        fig = plt.gcf()
                        fig.tight_layout()
                        for idx, ax in enumerate(fig.axes):
                            # Hide all the other axes...
                            if idx in (1, 2):
                                ax.yaxis.set_major_formatter(FormatStrFormatter('%.2f'))
                            ax._orig_position = ax.get_position()
                            ax.set_position([0.15, 0.11, 0.80, 0.79])
                            for axis in fig.axes:
                                if axis is not ax:
                                    axis.set_visible(False)

                            # fig.tight_layout()
                            plt.savefig(f'constrained_plots{dist_idx}\\{meas_config.replace(" ", "")}\\valves_{samples}\\constrained_noise_{noise}_MS{MS}_mode_{mode}_{idx+1}.eps')

                            ax.set_position(ax._orig_position)
                            for axis in fig.axes:
                                if axis is not ax:
                                    axis.set_visible(True)

                        plt.close()

                        def get_state_step(x_step, var):
                            var_step = [step[hex3_chen.x_vars.index(var)] for step in x_step]
                            return var_step

                        try:
                            # print(entry['data']['x_step'])
                            plt.plot(entry['data']['t_step'], get_state_step(entry['data']['x_step'], 'T1'),label='T1')
                            plt.plot(entry['data']['t_step'], get_state_step(entry['data']['x_step'], 'T2'),label='T2')
                            plt.plot(entry['data']['t_step'], get_state_step(entry['data']['x_step'], 'T3'),label='T3')
                            plt.ylabel('Temperature [C]')
                            plt.xlabel('Time [-]')
                            plt.ylim([130, 145])
                            plt.legend()
                            plt.hlines(hex3_chen.Ti_max, 0, 500, colors='black')
                            plt.savefig(f'constrained_plots{dist_idx}\\{meas_config.replace(" ", "")}\\temps_{samples}\\constrained_noise_{noise}_MS{MS}_mode_{mode}_temp.eps')
                            # plt.show()
                            plt.close()
                        except:
                            import traceback
                            traceback.print_exc()

                finally:
                    try:
                        controller.close()
                    except:
                        pass

    dataframe = pd.DataFrame.from_records(runs)
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    # pd.set_option('display.max_colwidth', -1)
    print(dataframe.sort_values(by='MS'))
    dataframe.to_pickle(f'constrained_plots{dist_idx}\\{meas_config.replace(" ", "")}\\constrained_{samples}.pkl')

    # code.interact('Interactive mode started... ', local=dict(globals(), **locals()))


def single_run(disturb_table, dist_idx, meas_config):
    mode = 'uc'
    # mode = 'uc'
    MS = -2
    noise = True

    hex3_chen.Ti_max = 135
    samples = 2500
    np.random.seed(21)
    try:
        # GP controller
        do_plot = True
        if mode == 'uc':
            from simulation.predictor_constrained_uopt import InputConstrainedController

            # train_data = f'optimal_u_constrained\\datasets\\test_u_constrained_prediction{12500}.csv'
            train_data = f'optimal_u_constrained\\datasets\\train_u_constrained_prediction{samples}.csv'
            controller = InputConstrainedController(train_data, meas_set=MS, noise=noise, meas_config=meas_config)
        elif mode == 'uc2':
            from simulation.predictor_constrained_uopt_v2 import InputConstrainedController

            # train_data = f'optimal_u_constrained\\datasets\\test_u_constrained_prediction{12500}.csv'
            train_data = f'optimal_u_constrained\\datasets\\train_u_constrained_prediction{samples}.csv'
            # train_data = f'optimal_u\\datasets\\train_u_prediction{500}.csv'
            controller = InputConstrainedController(train_data, meas_set=MS, noise=noise, meas_config=meas_config)
        elif mode == 'gc':
            from simulation.predictor_constrained_gradient import GradientConstrainedController

            # train_data = f'gradient_constrained\\datasets\\train_constrained_gradient{500}.csv'
            train_data = f'gradient\\datasets\\new_train_gradient{samples}.csv'
            controller = GradientConstrainedController(train_data, meas_set=MS, noise=noise)
        else:
            raise Exception(f'unknown mode {mode}')

        # total_loss, max_step_loss = main(MS, do_plot, controller, distrub_table, noise)
        try:
            result = main(MS, do_plot, controller, disturb_table, noise)
        except AssertionError:
            print(f'Mode {mode} | MS: {MS} | meas_config {meas_config} | noise: {noise}')
            plt.show()

        entry = {
            'data': result,
            'Normal loss': result['total loss'],
            'negative loss': result['negative loss'],
            # 'max step loss': result['max step loss'],
            'mode': mode,
            'MS': MS,
            'noise': noise,
        }
        print(entry['MS'], entry['mode'], entry['noise'])
        if do_plot:

            fig = plt.gcf()
            fig.tight_layout()
            for idx, ax in enumerate(fig.axes):
                # Hide all the other axes...
                if idx in (1, 2):
                    ax.yaxis.set_major_formatter(FormatStrFormatter('%.2f'))
                ax._orig_position = ax.get_position()
                ax.set_position([0.15, 0.11, 0.80, 0.79])
                for axis in fig.axes:
                    if axis is not ax:
                        axis.set_visible(False)

                # fig.tight_layout()
                plt.savefig(f'constrained_plots{dist_idx}\\{meas_config.replace(" ", "")}\\valves_{samples}\\constrained_noise_{noise}_MS{MS}_mode_{mode}_{idx+1}.eps')

                ax.set_position(ax._orig_position)
                for axis in fig.axes:
                    if axis is not ax:
                        axis.set_visible(True)

            plt.close()

        def get_state_step(x_step, var):
            var_step = [step[hex3_chen.x_vars.index(var)] for step in x_step]
            return var_step
        try:
            # print(entry['data']['x_step'])
            plt.plot(entry['data']['t_step'], get_state_step(entry['data']['x_step'], 'T1'),label='T1')
            plt.plot(entry['data']['t_step'], get_state_step(entry['data']['x_step'], 'T2'),label='T2')
            plt.plot(entry['data']['t_step'], get_state_step(entry['data']['x_step'], 'T3'),label='T3')
            plt.ylabel('Temperature [C]')
            plt.xlabel('Time [-]')
            plt.ylim([130, 145])
            plt.legend()
            plt.hlines(hex3_chen.Ti_max, 0, 500, colors='black')
            plt.savefig(f'constrained_plots{dist_idx}\\{meas_config.replace(" ", "")}\\temps_{samples}\\constrained_noise_{noise}_MS{MS}_mode_{mode}_temp.eps')
            # plt.show()
            plt.close()
        except:
            import traceback
            traceback.print_exc()

        # plt.plot(entry['data']['t_step'], entry['data']['x_step'])
        # plt.plot(entry['data']['t_step'], entry['data']['x_step'])

    finally:
        try:
            controller.close()
        except:
            pass
    import pprint
    import pandas as pd
    dataframe = pd.DataFrame.from_records([entry])
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    # pd.set_option('display.max_colwidth', -1)
    result = str(dataframe.sort_values(by='MS'))

    # dataframe.to_pickle(f'constrained_plots\\constrained_noise_{noise}_MS{MS}_mode_{mode}_temp.pkl')
    # code.interact('Interactive mode started... ', local=dict(globals(), **locals()))
    return result


if __name__ == '__main__':
    from disturbance_tables import constrained_distrub_table2, distrub_table1
    font = {'size': 14}
    matplotlib.rc('font', **font)
    disturb_table = distrub_table1

    results = []
    for meas_config in ('both', 't only', 'txv only'):
        for idx, dist_table in enumerate([distrub_table1, constrained_distrub_table2], start=1):
            r = single_run(dist_table, idx, meas_config)
            results.append((meas_config, idx, r))
            #full_run(dist_table, idx, meas_config)


    for meas_config, dist_set, data in results:
        print(f'Dist set {dist_set} | meas_config {meas_config}')
        print(data)

    # single_run(disturb_table)
