import numpy as np
from casadi import *

nlpopts = {'ipopt': {'print_level': 0}, 'print_time': False};
x_vars = ['alpha3', 'T', 'Tstar1', 'Tstar2', 'Tstar3', 'The1', 'The2', 'The3', 'Q1', 'Q2', 'Q3', 'Qloss1', 'Qloss2',
          'Qloss3', 'T1', 'T2', 'T3'];
u_vars = ['alpha1', 'alpha2']

meas_sets = {
    -4: ['T0', 'T', 'T1', 'T2', 'T3', 'Th1', 'Th2', 'Th3', 'alpha1', 'alpha2'],
    -3: ['T0', 'T', 'T1', 'T2', 'T3', 'The1', 'The2', 'The3', 'w0', 'wh1', 'wh2', 'wh3'],
    -2: ['T0', 'Th1', 'Th2', 'Th3', 'The1', 'The2', 'The3', 'T1', 'T2', 'T3'],
    -1: ['T0', 'T1', 'T2', 'T3', 'Th1', 'Th2', 'Th3'],
    1: ['T0', 'T1', 'T2', 'T3', 'Th1', 'Th2', 'Th3'],
    2: ['T0', 'Th1', 'Th2', 'Th3', 'The1', 'The2', 'The3'],
    3: ['T0', 'T', 'The1', 'The2', 'The3', 'w0', 'wh1', 'wh2', 'wh3'],
    4: ['T0', 'T', 'Th1', 'Th2', 'Th3', 'alpha1', 'alpha2']
}

Ti_max = 1500


def model(par):
    T = SX.sym('T');
    Tstar1 = SX.sym('Tstar1');
    Tstar2 = SX.sym('Tstar2');
    Tstar3 = SX.sym('Tstar3');
    The1 = SX.sym('The1');
    The2 = SX.sym('The2');
    The3 = SX.sym('The3');
    Q1 = SX.sym('Q1');
    Q2 = SX.sym('Q2');
    Q3 = SX.sym('Q3');
    Qloss1 = SX.sym('Qloss1');
    Qloss2 = SX.sym('Qloss2');
    Qloss3 = SX.sym('Qloss3');
    T1 = SX.sym('T1');
    T2 = SX.sym('T2');
    T3 = SX.sym('T3');
    alpha1 = SX.sym('alpha1');
    alpha2 = SX.sym('alpha2');
    alpha3 = SX.sym('alpha3');

    T0 = par['T0'];
    w0 = par['w0'];
    Th1 = par['Th1'];
    Th2 = par['Th2'];
    Th3 = par['Th3'];
    wh1 = par['wh1'];
    wh2 = par['wh2'];
    wh3 = par['wh3'];
    UA1 = par['UA1'];
    UA2 = par['UA2'];
    UA3 = par['UA3'];

    Ts = par['Ts'];
    h1 = par['h1'];
    h2 = par['h2'];
    h3 = par['h3'];

    dTlm1 = ((Th1 - Tstar1) * (The1 - T0) * ((Th1 - Tstar1) + (The1 - T0)) / 2) ** (1 / 3);
    dTlm2 = ((Th2 - Tstar2) * (The2 - T0) * ((Th2 - Tstar2) + (The2 - T0)) / 2) ** (1 / 3);
    dTlm3 = ((Th3 - Tstar3) * (The3 - T0) * ((Th3 - Tstar3) + (The3 - T0)) / 2) ** (1 / 3);

    f0 = - T + alpha1 * T1 + alpha2 * T2 + alpha3 * T3;
    f01 = alpha1 + alpha2 + alpha3 - 1;
    f11 = - Q1 + w0 * alpha1 * (Tstar1 - T0);
    f12 = - Q2 + w0 * alpha2 * (Tstar2 - T0);
    f13 = - Q3 + w0 * alpha3 * (Tstar3 - T0);
    f21 = - Q1 + UA1 * dTlm1;
    f22 = - Q2 + UA2 * dTlm2;
    f23 = - Q3 + UA3 * dTlm3;
    f31 = - Q1 + wh1 * (Th1 - The1);
    f32 = - Q2 + wh2 * (Th2 - The2);
    f33 = - Q3 + wh3 * (Th3 - The3);
    f41 = - Qloss1 + w0 * alpha1 * (T1 - Tstar1);
    f42 = - Qloss2 + w0 * alpha2 * (T2 - Tstar2);
    f43 = - Qloss3 + w0 * alpha3 * (T3 - Tstar3);
    f51 = - Qloss1 + h1 * (Ts - T1);
    f52 = - Qloss2 + h2 * (Ts - T2);
    f53 = - Qloss3 + h3 * (Ts - T3);

    x = vertcat(alpha3, T, Tstar1, Tstar2, Tstar3, The1, The2, The3, Q1, Q2, Q3, Qloss1, Qloss2, Qloss3, T1, T2, T3);
    f = vertcat(f0, f01, f11, f12, f13, f21, f22, f23, f31, f32, f33, f41, f42, f43, f51, f52, f53);
    u = vertcat(alpha1, alpha2);
    J = -T;

    return {'x': x, 'u': u, 'f': f, 'J': J}


def output(u, par, x0=None):
    m = model(par);
    nx = np.prod(m['x'].shape);
    nu = np.prod(m['u'].shape);
    nf = np.prod(m['f'].shape);

    if x0 is None:
        x0 = np.array([0.33] * (nu + 1) + [(par['T0'] + par['Th1']) / 2, (par['T0'] + par['Th2']) / 2,
                                           (par['T0'] + par['Th3']) / 2] * 2 + [par['T0']] * (nx - 2 * (nu + 1) - 1))

    nlp = {}  # NLP declaration
    nlp['x'] = vertcat(m['u'], m['x'])  # decision vars
    nlp['f'] = m['J']  # objective
    nlp['g'] = m['f']  # constraints

    # Create solver instance
    F = nlpsol('F', 'ipopt', nlp, nlpopts);

    # Solve the problem using a guess
    lbx = np.array([*u] + [0] * (1) + [-inf] * (nx - 1));  # constraint on inputs and first state (last flow split)
    ubx = np.array([*u] + [1] * (1) + [+inf] * (
            nx - 1));  # upper limit on splits is not necessary, but will automatically be satisfied

    lbx[(nu + 2):(nu + 2 + 3 * 2)] = par['T0'];  # constraint on temperatures
    ubx[(nu + 2):(nu + 2 + 3 * 2)] = np.array(
        [par['Th1'], par['Th2'], par['Th3'], par['Th1'], par['Th2'], par['Th3']]);  # constraint on temperatures

    r = F(x0=x0, lbg=np.zeros(nf), ubg=np.zeros(nf), lbx=lbx, ubx=ubx);

    sol = r['x'].full().reshape(-1)

    return {'x': sol[-nx:], 'success': F.stats()['success']}


def cost(u, par):
    m = model(par);
    F = Function('F', [m['x'], m['u']], [m['J']], ['x', 'u'], ['J']);
    out = output(u, par);  # np.zeros(nx);
    J = F(out['x'], u);
    return {'J': J.full().reshape(-1), 'success': out['success']};


def grad(u, par):
    m = model(par);

    F = Function('F', [m['x'], m['u']], [m['f'], m['J']], ['x', 'u'], ['f', 'J']);
    G = rootfinder('G', 'newton', F)
    out = output(u, par);  # np.zeros(nx);
    Jufun = G.factory('Ju', ['x', 'u'], ['jac:J:u']);

    delta = 0;
    Ju = Jufun(out['x'] + delta, u).full().reshape(-1)
    # while not G.stats()['success']:
    #     delta = delta*10;
    #     Ju = Jufun(xguess+delta, u).full().reshape(-1)
    return {'grad': Ju, 'success': True};


def output_meas(meas_set, u, par):
    meas_vars = meas_sets[meas_set];
    y = np.zeros((len(meas_vars),));
    out = output(u, par);
    x = out['x'];
    for i, var in enumerate(meas_vars):
        if var in par:
            y[i] = par[var];
        elif var in u_vars:
            y[i] = u[u_vars.index(var)];
        elif var in x_vars:
            y[i] = x[x_vars.index(var)];
        else:
            y[i] = np.nan;
    return {'x': x, 'par': par, 'y': y, 'T': x[x_vars.index('T')], 'success': out['success']}


def optim(par, x0=None):
    m = model(par);
    nx = np.prod(m['x'].shape);
    nu = np.prod(m['u'].shape);
    nf = np.prod(m['f'].shape);

    nlp = {}  # NLP declaration
    nlp['x'] = vertcat(m['u'], m['x'])  # decision vars
    nlp['f'] = m['J']  # objective
    nlp['g'] = m['f']  # constraints

    # Create solver instance
    F = nlpsol('F', 'ipopt', nlp, nlpopts);

    Tbackoff = 1;
    # Trand = 1;
    alphabackoff = 1e-3;

    if x0 is None:
        # x0 = np.zeros(nx+nu);
        # x0[:nu+1] = 1/(nu+1);
        # x0[(nu+2):(nu+2+3*2)] = par['T0'] + Tbackoff #+ Trand*np.random.rand(3*2); # [Tstar1,Tstar2,Tstar3,The1,The2,The3]
        x0 = np.array([0.33] * (nu + 1) + [(par['T0'] + par['Th1']) / 2, (par['T0'] + par['Th2']) / 2,
                                           (par['T0'] + par['Th3']) / 2] * 2 + [par['T0']] * (nx - 2 * (nu + 1) - 1))

    # Solve the problem using first guess

    lbx = np.array(
        [alphabackoff] * (nu + 1) + [-inf] * (nx - 1));  # constraint on inputs and first state (last flow split)
    ubx = np.array([1] * (nu + 1) + [+inf] * (
            nx - 1));  # upper limit on splits is not necessary, but will automatically be satisfied

    lbx[(nu + 2):(nu + 2 + 3 * 2)] = par['T0'];  # constraint on temperatures
    ubx[(nu + 2):(nu + 2 + 3)] = np.array(
        [min(Ti_max, t) for t in [par['Th1'], par['Th2'], par['Th3']]]);  # constraint on temperatures
    ubx[(nu + 2 + 3):(nu + 2 + 3 * 2)] = np.array([par['Th1'], par['Th2'], par['Th3']]);  # constraint on temperatures

    r = F(x0=x0, lbg=np.zeros(nf), ubg=np.zeros(nf), lbx=lbx, ubx=ubx);

    # while not F.stats()['success']:
    #     Trand = Trand + 1;
    #     x0[(nu+2):(nu+2+3*2)] = par['T0'] + Tbackoff + Trand*np.random.rand(3*2);
    #     r = F(x0=x0, lbg=np.zeros(nf), ubg=np.zeros(nf), lbx=lbx, ubx=ubx);

    sol = r['x'].full().reshape(-1)

    return {'u': sol[:nu], 'x': sol[-nx:], 'success': F.stats()['success']}
