import code
import sys

import matplotlib
import numpy as np
import pandas
from matplotlib.ticker import FormatStrFormatter

import hex3_chen_old as hex3_chen

matplotlib.use('Qt5agg')
import matplotlib.pyplot as plt

from simulation.predictor_gradient import GradientController
from simulation.predictor_uopt import InputController
from simulation.simulation import main


def full_run(disturb_table, dist_idx):
    runs = []
    import pandas as pd
    samples = 500
    for mode in ['g', 'u']:
        for MS in range(1, 5):
            for noise in [False, True]:
                try:
                    # GP controller
                    np.random.seed(21)
                    do_plot = True

                    if mode == 'u':
                        train_data = f'optimal_u\\datasets\\train_u_prediction{samples}.csv'
                        controller = InputController(train_data, meas_set=MS, noise=noise)
                    elif mode == 'g':
                        train_data = f'gradient\\datasets\\new_train_gradient{samples}.csv'
                        controller = GradientController(train_data, meas_set=MS, noise=noise)
                    else:
                        raise Exception(f'unknown mode {mode}')

                    result = main(MS, do_plot, controller, disturb_table, noise)
                    entry = {
                        'data': result,
                        'total loss': result['total loss'],
                        'negative loss': result['negative loss'],
                        # 'max step loss': result['max step loss'],
                        'mode': mode,
                        'MS': MS,
                        'noise': noise,
                    }
                    runs.append(entry)
                    if do_plot:

                        fig = plt.gcf()
                        fig.tight_layout()
                        for idx, ax in enumerate(fig.axes):
                            # Hide all the other axes...
                            if idx in (1, 2):
                                ax.yaxis.set_major_formatter(FormatStrFormatter('%.2f'))
                            ax._orig_position = ax.get_position()
                            ax.set_position([0.15, 0.11, 0.80, 0.79])
                            for axis in fig.axes:
                                if axis is not ax:
                                    axis.set_visible(False)

                            # fig.tight_layout()
                            # if idx in (1, 2):
                            #     plt.savefig(f'unconstrained_plots{dist_idx}\\gradients_{samples}\\unconstrained_noise_{noise}_MS{MS}_mode_{mode}_{idx+1}.eps')
                            plt.savefig(f'unconstrained_plots{dist_idx}\\valves_{samples}\\unconstrained_noise_{noise}_MS{MS}_mode_{mode}_{idx+1}.eps')

                            ax.set_position(ax._orig_position)
                            for axis in fig.axes:
                                if axis is not ax:
                                    axis.set_visible(True)

                        plt.close()
                finally:
                    try:
                        controller.close()
                    except:
                        pass

    dataframe = pd.DataFrame.from_records(runs)
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    # pd.set_option('display.max_colwidth', -1)
    print(dataframe.sort_values(by='MS'))
    # code.interact('Interactive mode started... ', local=dict(globals(), **locals()))


def single_run(disturb_table, dist_idx):
    # mode = 'g'
    mode = 'u'
    MS = 1
    noise = False
    samples = 500
    results = {}
    try:
        # GP controller
        np.random.seed(21)
        do_plot = True
        if mode == 'u':
            train_data = f'optimal_u\\datasets\\train_u_prediction{samples}.csv'
            controller = InputController(train_data, meas_set=MS, noise=noise)

        elif mode == 'g':
            train_data = f'gradient\\datasets\\new_train_gradient{samples}.csv'
            controller = GradientController(train_data, meas_set=MS, noise=noise)

        else:
            raise Exception(f'unknown mode {mode}')

        # total_loss, max_step_loss = main(MS, do_plot, controller, distrub_table, noise)
        try:
            result = main(MS, do_plot, controller, disturb_table, noise)
        except AssertionError:
            print(f'Mode {mode} | MS: {MS} | noise: {noise}')
            plt.show()
            raise

        entry = {
            'data': result,
            'total loss': result['total loss'],
            'negative loss': result['negative loss'],
            # 'max step loss': result['max step loss'],
            'mode': mode,
            'MS': MS,
            'noise': noise,
        }

        if do_plot:

            fig = plt.gcf()
            fig.tight_layout()
            for idx, ax in enumerate(fig.axes):
                # Hide all the other axes...
                if idx in (1, 2):
                    ax.yaxis.set_major_formatter(FormatStrFormatter('%.2f'))
                ax._orig_position = ax.get_position()
                ax.set_position([0.15, 0.11, 0.80, 0.79])
                for axis in fig.axes:
                    if axis is not ax:
                        axis.set_visible(False)

                # fig.tight_layout()
                # if idx in (1, 2):
                #     plt.savefig(f'unconstrained_plots{dist_idx}\\gradients_{samples}\\unconstrained_noise_{noise}_MS{MS}_mode_{mode}_{idx+1}.eps')

                plt.savefig(f'unconstrained_plots{dist_idx}\\diverging_unconstrained_noise_{noise}_MS{MS}_mode_{mode}_{idx+1}.eps')

                # if idx in (1, 2):
                #     plt.savefig(f'unconstrained_plots{dist_idx}\\only_optimal_unconstrained_noise_{noise}_MS{MS}_mode_{mode}_{idx+1}.eps')

                ax.set_position(ax._orig_position)
                for axis in fig.axes:
                    if axis is not ax:
                        axis.set_visible(True)

            # plt.savefig(f'unconstrained_MS{MS}_mode_{mode}_{}.png', dpi=400)
            plt.show()

        # def get_state_step(x_step, var):
        #     var_step = [step[hex3_chen.x_vars.index(var)] for step in x_step]
        #     return var_step
        # try:
        #     # print(entry['data']['x_step'])
        #     plt.plot(entry['data']['t_step'], get_state_step(entry['data']['x_step'], 'T1'),label='T1')
        #     plt.plot(entry['data']['t_step'], get_state_step(entry['data']['x_step'], 'T2'),label='T2')
        #     plt.plot(entry['data']['t_step'], get_state_step(entry['data']['x_step'], 'T3'),label='T3')
        #     plt.ylim([130, 145])
        #     plt.legend()
        #     plt.show()
        # except:
        #     import traceback
        #     traceback.print_exc()

        # plt.plot(entry['data']['t_step'], entry['data']['x_step'])
        # plt.plot(entry['data']['t_step'], entry['data']['x_step'])

    finally:
        try:
            controller.close()
        except:
            pass
    import pprint
    import pandas as pd
    dataframe = pd.DataFrame.from_records([entry])
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    # pd.set_option('display.max_colwidth', -1)

    def to_ltx(dataframe: pandas.DataFrame):
        pd_c = dataframe.copy()
        del pd_c['data']
        print(pd_c.to_latex())

    print(dataframe.sort_values(by='MS'))
    code.interact('Interactive mode started... ', local=dict(globals(), **locals()))


if __name__ == '__main__':
    from disturbance_tables import distrub_table2, distrub_table1
    font = {'size': 14}
    matplotlib.rc('font', **font)
    disturb_table = distrub_table1

    # for idx, dist_table in enumerate([distrub_table1, distrub_table2], start=1):
    #     full_run(dist_table, idx)
    #     break

    # full_run(disturb_table)
    #
    single_run(disturb_table, 1)
